// SPDX-FileCopyrightText: 2022 Matteo Settenvini <matteo.settenvini@montecristosoftware.eu>
// SPDX-License-Identifier: AGPL-3.0-or-later

mod api_client;
mod config;
mod constants;
mod recipe;

use {
    crate::api_client::ApiClient,
    crate::config::Config,
    crate::recipe::Recipe,
    clap::{arg, command, ArgMatches, Command},
    reqwest::{StatusCode, Url},
    std::path::PathBuf,
};

#[tokio::main(flavor = "multi_thread")]
async fn main() -> anyhow::Result<()> {
    let args = parse_args();
    let mut configuration = Config::new();

    match args.subcommand() {
        Some(("init", sub_matches)) => {
            let server = sub_matches
                .get_one::<String>("server")
                .expect("Mandatory parameter <server>");
            tokio::task::block_in_place(move || -> anyhow::Result<()> {
                configuration
                    .credentials
                    .add(Url::parse(server)?)
                    .expect("Unable to authenticate to NextCloud instance");
                Ok(())
            })?;
        }
        Some(("import", sub_matches)) => {
            let api_client = get_api_client(&sub_matches, &configuration)?;
            for url in sub_matches
                .get_many::<String>("url")
                .expect("At least one url is required")
            {
                let response = api_client
                    .client
                    .post(api_client.base_url.join("apps/cookbook/import")?)
                    .json(&serde_json::json!({
                        "url": url,
                    }))
                    .send()
                    .await?;
                println!("{:#?}", response); // TODO
                assert!([StatusCode::OK, StatusCode::CONFLICT].contains(&response.status()));
            }
        }
        Some(("schedule", sub_matches)) => {
            let api_client = get_api_client(&sub_matches, &configuration)?;
            let recipes = api_client
                .client
                .get(api_client.base_url.join("apps/cookbook/api/recipes")?)
                .send()
                .await?;
            println!("{:#?}", recipes.json::<Vec<Recipe>>().await?); // TODO
        }
        Some(("schedule-csv", sub_matches)) => {
            let csv_file = sub_matches
                .get_one::<PathBuf>("csv_file")
                .expect("<csv_file> is a mandatory parameter, it cannot be missing");
            let calendar_name = sub_matches
                .get_one::<String>("calendar_name")
                .expect("<calendar_name> is a mandatory parameter, it cannot be missing");

            let mut csv = csv::Reader::from_path(csv_file)?;

            #[derive(serde::Deserialize)]
            struct CsvRecord {
                day: chrono::naive::NaiveDate,
                lunch: String,
                dinner: String,
            }

            let recipe_urls = csv.deserialize::<CsvRecord>().fold(
                std::collections::HashSet::new(),
                |mut set, r| {
                    if let Ok(r) = r {
                        set.insert(r.lunch);
                        set.insert(r.dinner);
                    }
                    set
                },
            );

            let api_client = get_api_client(&sub_matches, &configuration)?;
            for url in recipe_urls {
                let response = api_client
                    .client
                    .post(api_client.base_url.join("apps/cookbook/import")?)
                    .json(&serde_json::json!({
                        "url": url,
                    }))
                    .send()
                    .await?;
                println!("{:#?}", response); // TODO
                assert!([StatusCode::OK, StatusCode::CONFLICT].contains(&response.status()));
            }
        }
        _ => unreachable!("Exhausted list of subcommands and subcommand_required prevents `None`"),
    };

    Ok(())
}

fn parse_args() -> ArgMatches {
    let server_arg = arg!(-s --server <server> "NextCloud server to connect to").required(false);

    command!()
        .propagate_version(true)
        .subcommand_required(true)
        .arg_required_else_help(true)
        .subcommand(
            Command::new("init")
                .about("Authenticate against the provided NextCloud server")
                .arg(arg!(<server> "NextCloud server to connect to")),
        )
        .subcommand(
            Command::new("import")
                .about("Import the given URLs into NextCloud's cookbook")
                .arg(server_arg.clone())
                .arg(arg!(<url> ... "One or more URLs each pointing to page with a recipe to import in NextCloud")),
        )
        .subcommand(
            Command::new("schedule")
                .about("")
                .arg(server_arg.clone())
                .arg(arg!(-d --days <days> "")
                    .value_parser(clap::builder::RangedU64ValueParser::<u32>::new().range(1..))
                    .required(false)
                    .default_value("7"))
        )
        .subcommand(
            Command::new("schedule-csv")
                .about("TEMPORARY WIP FUNCTION, UNSTABLE")
                .arg(server_arg.clone())
                .arg(arg!(<calendar_name> ""))
                .arg(arg!(<csv_file> "").value_parser(clap::value_parser!(PathBuf)))
        )
        .get_matches()
}

fn get_api_client(sub_matches: &ArgMatches, configuration: &Config) -> anyhow::Result<ApiClient> {
    let server_name = sub_matches
        .get_one::<String>("server")
        .unwrap_or_else(|| {
            let servers = &configuration.credentials.servers;
            match servers.len() {
                0 => panic!("No NextCloud server set up yet, use '{} init' first", env!("CARGO_BIN_NAME")),
                1 => servers.iter().next().unwrap().0,
                _ => panic!("More than one NextCloud server set up, use the '--server' option to specify which one to use. Known servers: {:#?}", servers.keys().collect::<Vec<_>>()),
            }
        });

    ApiClient::new(&server_name, &configuration)
}
